﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace XamDataChartNoiseDataDemo
{
    public class PidDataSet
    {
        //initial values
        private double minDataVal = 0;
        private double maxDataVal = 0;
        private double minDispVal = 0;
        private double maxDispVal = 0;

        public string Name { get; private set; }
        public int Multiplier { get; private set; }
        public CategoryDataCollection PlotData2 { get; private set; }
        public double MinDataVal { get { return minDataVal; } }
        public double MaxDataVal { get { return maxDataVal; } }
        public double MinDispVal { get { return minDispVal; } set { minDispVal = value; } }
        public double MaxDispVal { get { return maxDispVal; } set { maxDispVal = value; } }

        //public SciChart.Charting.Model.DataSeries.XyDataSeries<int, double> DataSeries { get; set; }


        // class defining complete set of data for a single pid
        public PidDataSet(string name, IList<double> framedata, double PlotMin, double PlotMax, int multiplier)
        {
            //Create new collection of data for this item
            PlotData2 = new CategoryDataCollection();
            Name = name;
            Multiplier = multiplier;

            //loop thru all the frames getting the data and computing min/max
            //NaN is added to plot data but not used in min/max computation
            minDataVal = double.MaxValue;
            maxDataVal = double.MinValue;
            for (int i = 0; i < framedata.Count; i++)
            {
                double val = framedata[i];
                PlotData2.Add(new CategoryDataPoint { Index = i, Value = val, Category = i.ToString() });

                if (double.IsNaN(val) == true)
                    continue;

                minDataVal = Math.Min(minDataVal, val);
                maxDataVal = Math.Max(maxDataVal, val);

            }

            //assign the default scaling for this current scale item
            MinDispVal = PlotMin;
            MaxDispVal = PlotMax;

            //for performance testing, add duplicate data set to increase amount of data
            //DataSeries = new SciChart.Charting.Model.DataSeries.XyDataSeries<int, double>();
            if (Multiplier > 0)
            {
                for (int j = 0; j < Multiplier; j++)
                {
                    for (int i = 0; i < framedata.Count; i++)
                    {
                        double val = framedata[i];
                        int idx = PlotData2.Count;
                        PlotData2.Add(new CategoryDataPoint { Index = idx, Value = val, Category = idx.ToString() });
                        //DataSeries.Append(idx, val);
                    }
                }
            }
        }
    }



    #region Infragistics data Classes

    public class CategoryDataCollection : ObservableCollection<CategoryDataPoint>
    {
    }
    public class CategoryDataPoint
    {
        public CategoryDataPoint() { }

        public int Index { get; set; }
        public string Category { get; set; }
        public double Value { get; set; }
        //public double High { get; set; }
        //public double Low { get; set; }
    }
    #endregion
}
