﻿using Infragistics;
using Infragistics.Controls.Charts;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Media;

namespace XamDataChartNoiseDataDemo.ImmediateLineSeries
{
    public class ImmediateLineSeries : Series
    {
        public ImmediateLineSeries()
        {
            PreviousFrame = new OrderedScatterFrame();
            CurrentFrame = new OrderedScatterFrame();
            TransitionFrame = new OrderedScatterFrame();
            ThumbnailFrame = new OrderedScatterFrame();

            DefaultStyleKey = typeof(ImmediateLineSeries);
        }

        protected override SeriesView CreateView()
        {
            return new ImmediateLineSeriesWBXView(this);
        }

        protected override void OnViewCreated(SeriesView view)
        {
            base.OnViewCreated(view);

            ImmediateView = (ImmediateLineSeriesBaseView)view;
        }

        #region Protected Properties
        protected ImmediateLineSeriesBaseView ImmediateView { get; set; }
        protected ImmediateLineSeriesBaseView ThumbView { get; set; }

        protected OrderedScatterFrame PreviousFrame { get; set; }
        protected OrderedScatterFrame CurrentFrame { get; set; }
        protected OrderedScatterFrame TransitionFrame { get; set; }
        protected OrderedScatterFrame ThumbnailFrame { get; set; }
        #endregion

        #region XAxis Dependency Property
        /// <summary>
        /// Gets or sets the effective x-axis for the current object.
        /// <para>
        /// This is a dependency property.
        /// </para>
        /// </summary>
        public NumericXAxis XAxis
        {
            get
            {
                return (NumericXAxis)GetValue(XAxisProperty);
            }
            set
            {
                SetValue(XAxisProperty, value);
            }
        }

        internal const string XAxisPropertyName = "XAxis";

        /// <summary>
        /// Identifies the XAxis dependency property.
        /// </summary>
        public static readonly DependencyProperty XAxisProperty = DependencyProperty.Register(
            XAxisPropertyName, typeof(NumericXAxis),
            typeof(ImmediateLineSeries),
            new PropertyMetadata(null,
                delegate (DependencyObject sender,
                DependencyPropertyChangedEventArgs e)
                {
                    (sender as ImmediateLineSeries).RaisePropertyChanged(
                        XAxisPropertyName,
                        e.OldValue, e.NewValue);
                }));
        #endregion

        #region YAxis Depedency Property
        /// <summary>
        /// Gets or sets the effective y-axis for the current object.
        /// <para>
        /// This is a dependency property.
        /// </para>
        /// </summary>
        public NumericYAxis YAxis
        {
            get
            {
                return (NumericYAxis)GetValue(YAxisProperty);
            }
            set
            {
                SetValue(YAxisProperty, value);
            }
        }

        internal const string YAxisPropertyName = "YAxis";

        /// <summary>
        /// Identifies the YAxis dependency property.
        /// </summary>
        public static readonly DependencyProperty YAxisProperty = DependencyProperty.Register(
            YAxisPropertyName,
            typeof(NumericYAxis), typeof(ImmediateLineSeries),
            new PropertyMetadata(
                null,
                delegate (DependencyObject sender,
                DependencyPropertyChangedEventArgs e)
                {
                    (sender as ImmediateLineSeries).RaisePropertyChanged(
                        YAxisPropertyName,
                        e.OldValue, e.NewValue);
                }));
        #endregion

        #region XMemberPath Dependency Property and XColumn Property
        /// <summary>
        /// Gets or sets the value mapping property for the current series object.
        /// <para>
        /// This is a dependency property.
        /// </para>
        /// </summary>
        public string XMemberPath
        {
            get
            {
                return (string)GetValue(XMemberPathProperty);
            }
            set
            {
                SetValue(XMemberPathProperty, value);
            }
        }

        internal const string XMemberPathPropertyName = "XMemberPath";

        /// <summary>
        /// Identifies the XMemberPath dependency property.
        /// </summary>
        public static readonly DependencyProperty XMemberPathProperty = DependencyProperty.Register(
            XMemberPathPropertyName, typeof(string),
            typeof(ImmediateLineSeries),
            new PropertyMetadata(null,
                delegate (DependencyObject sender,
                DependencyPropertyChangedEventArgs e)
                {
                    (sender as ImmediateLineSeries).RaisePropertyChanged(
                        XMemberPathPropertyName,
                        e.OldValue, e.NewValue);
                }));
        /// <summary>
        /// The IFastItemColumn containing X values.
        /// </summary>
        protected internal IFastItemColumn<double> XColumn
        {
            get { return xColumn; }
            private set
            {
                if (xColumn != value)
                {
                    IFastItemColumn<double> oldXColumn = XColumn;

                    xColumn = value;
                    RaisePropertyChanged(XColumnPropertyName, oldXColumn, XColumn);
                }
            }
        }
        private IFastItemColumn<double> xColumn;
        internal const string XColumnPropertyName = "XColumn";
        #endregion

        #region YMemberPath Dependency Property and YColumn Property
        /// <summary>
        /// Gets or sets the value mapping property for the current series object.
        /// <para>
        /// This is a dependency property.
        /// </para>
        /// </summary>
        public string YMemberPath
        {
            get
            {
                return (string)GetValue(YMemberPathProperty);
            }
            set
            {
                SetValue(YMemberPathProperty, value);
            }
        }

        protected override void OnRender(DrawingContext drawingContext)
        {
            base.OnRender(drawingContext);

            ImmediateView.OnRender(drawingContext);
        }



        internal const string YMemberPathPropertyName = "YMemberPath";

        /// <summary>
        /// Identifies the YMemberPath dependency property.
        /// </summary>
        public static readonly DependencyProperty YMemberPathProperty = DependencyProperty.Register(
            YMemberPathPropertyName, typeof(string),
            typeof(ImmediateLineSeries),
            new PropertyMetadata(null,
                delegate (DependencyObject sender, DependencyPropertyChangedEventArgs e)
                {
                    (sender as ImmediateLineSeries).RaisePropertyChanged(
                        YMemberPathPropertyName,
                        e.OldValue, e.NewValue);
                }));

        #region UseHairline Dependency Property and YColumn Property
        /// <summary>
        /// Gets or sets whether a hairline polyline should be used to render the series rather than a min/max range.
        /// <para>
        /// This is a dependency property.
        /// </para>
        /// </summary>
        public bool UseHairline
        {
            get
            {
                return (bool)GetValue(UseHairlineProperty);
            }
            set
            {
                SetValue(UseHairlineProperty, value);
            }
        }

        internal const string UseHairlinePropertyName = "UseHairline";

        /// <summary>
        /// Identifies the UseHairline dependency property.
        /// </summary>
        public static readonly DependencyProperty UseHairlineProperty = DependencyProperty.Register(
            UseHairlinePropertyName, typeof(bool),
            typeof(ImmediateLineSeries),
            new PropertyMetadata(false,
                delegate (DependencyObject sender, DependencyPropertyChangedEventArgs e)
                {
                    (sender as ImmediateLineSeries).RaisePropertyChanged(
                        UseHairlinePropertyName,
                        e.OldValue, e.NewValue);
                }));
        #endregion


        #region IsAntialiasingEnabled Dependency Property and YColumn Property
        /// <summary>
        /// Gets or sets whether a polyline should use anti aliasing for the stroke if in WBX mode.
        /// <para>
        /// This is a dependency property.
        /// </para>
        /// </summary>
        public bool IsAntialiasingEnabled
        {
            get
            {
                return (bool)GetValue(IsAntialiasingEnabledProperty);
            }
            set
            {
                SetValue(IsAntialiasingEnabledProperty, value);
            }
        }

        internal const string IsAntialiasingEnabledPropertyName = "IsAntialiasingEnabled";

        /// <summary>
        /// Identifies the IsAntialiasingEnabled dependency property.
        /// </summary>
        public static readonly DependencyProperty IsAntialiasingEnabledProperty = DependencyProperty.Register(
            IsAntialiasingEnabledPropertyName, typeof(bool),
            typeof(ImmediateLineSeries),
            new PropertyMetadata(true,
                delegate (DependencyObject sender, DependencyPropertyChangedEventArgs e)
                {
                    (sender as ImmediateLineSeries).RaisePropertyChanged(
                        IsAntialiasingEnabledPropertyName,
                        e.OldValue, e.NewValue);
                }));
        #endregion

        #region UpscaleOPD Dependency Property
        /// <summary>
        /// Gets or sets whether the OPD thumbnail should be upscaled to the main view dimensions.  This allows a more
        /// accurate shape to be rendered in the OPD but may produce various aliasing artifacts and lower visual quality.
        /// </summary>
        public bool UpscaleOPD
        {
            get { return (bool)GetValue(UpscaleOPDProperty); }
            set { SetValue(UpscaleOPDProperty, value); }
        }
        internal const string UpscaleOPDPropertyName = "UpscaleOPD";

        // Using a DependencyProperty as the backing store for UpscaleOPD.  This enables animation, styling, binding, etc...
        public static readonly DependencyProperty UpscaleOPDProperty =
            DependencyProperty.Register(UpscaleOPDPropertyName, typeof(bool), typeof(ImmediateLineSeries),
                new PropertyMetadata(false,
                    delegate (DependencyObject sender, DependencyPropertyChangedEventArgs e)
                    {
                        (sender as ImmediateLineSeries).RaisePropertyChanged(UpscaleOPDPropertyName, e.OldValue, e.NewValue);
                    }));
        #endregion

        /// <summary>
        /// The IFastItemColumn containing Y values.
        /// </summary>
        protected internal IFastItemColumn<double> YColumn
        {
            get { return yColumn; }
            private set
            {
                if (yColumn != value)
                {
                    IFastItemColumn<double> oldYColumn = YColumn;

                    yColumn = value;
                    RaisePropertyChanged(YColumnPropertyName, oldYColumn, YColumn);
                }
            }
        }
        private IFastItemColumn<double> yColumn;
        internal const string YColumnPropertyName = "YColumn";
        #endregion

        protected override void PropertyUpdatedOverride(
            object sender,
            string propertyName,
            object oldValue,
            object newValue)
        {
            base.PropertyUpdatedOverride(sender, propertyName, oldValue, newValue);

            switch (propertyName)
            {
                case FastItemsSourcePropertyName:
                    if (oldValue as FastItemsSource != null)
                    {
                        (oldValue as FastItemsSource).DeregisterColumn(XColumn);
                        (oldValue as FastItemsSource).DeregisterColumn(YColumn);
                        XColumn = null;
                        YColumn = null;
                    }

                    if (newValue as FastItemsSource != null)
                    {
                        XColumn = RegisterDoubleColumn(XMemberPath);
                        YColumn = RegisterDoubleColumn(YMemberPath);
                    }

                    if ((YAxis != null && !YAxis.UpdateRange()) ||
                        (XAxis != null && !XAxis.UpdateRange()))
                    {
                        RenderSeries(false);
                    }

                    break;

                case XAxisPropertyName:
                    if (oldValue != null)
                    {
                        (oldValue as Axis).DeregisterSeries(this);
                    }

                    if (newValue != null)
                    {
                        (newValue as Axis).RegisterSeries(this);
                    }

                    if ((XAxis != null && !XAxis.UpdateRange()) ||
                        (newValue == null && oldValue != null))
                    {
                        RenderSeries(false);
                    }
                    this.NotifyThumbnailAppearanceChanged();
                    break;

                case YAxisPropertyName:
                    if (oldValue != null)
                    {
                        (oldValue as Axis).DeregisterSeries(this);
                    }

                    if (newValue != null)
                    {
                        (newValue as Axis).RegisterSeries(this);
                    }
                    if ((YAxis != null && !YAxis.UpdateRange()) ||
                        (newValue == null && oldValue != null))
                    {
                        RenderSeries(false);
                    }
                    this.NotifyThumbnailAppearanceChanged();
                    break;

                #region X Mapping and Column
                case XMemberPathPropertyName:
                    if (FastItemsSource != null)
                    {
                        FastItemsSource.DeregisterColumn(XColumn);
                        XColumn = RegisterDoubleColumn(XMemberPath);
                    }

                    break;

                case XColumnPropertyName:
                    if (XAxis != null && !XAxis.UpdateRange())
                    {
                        RenderSeries(false);
                    }
                    break;
                #endregion

                #region Y Mapping and Column
                case YMemberPathPropertyName:
                    if (FastItemsSource != null)
                    {
                        FastItemsSource.DeregisterColumn(YColumn);
                        YColumn = RegisterDoubleColumn(YMemberPath);
                    }

                    break;

                case YColumnPropertyName:
                    if (YAxis != null && !YAxis.UpdateRange())
                    {
                        RenderSeries(false);
                    }
                    break;
                #endregion
                case UseHairlinePropertyName:
                case IsAntialiasingEnabledPropertyName:
                case UpscaleOPDPropertyName:
                    RenderSeries(false);
                    break;
            }
        }

        /// <summary>
        /// Renders the series.
        /// </summary>
        /// <param name="animate">True if the change should be animated.</param>
        protected override void RenderSeriesOverride(bool animate)
        {
            if (ClearAndAbortIfInvalid(View))
            {
                return;
            }
            //TODO: support animation

            PrepareFrame(CurrentFrame, ImmediateView);
            RenderFrame(CurrentFrame, ImmediateView);
        }

        protected override void RenderThumbnail(Rect viewportRect, RenderSurface surface)
        {
            base.RenderThumbnail(viewportRect, surface);

            if (ThumbnailView == null || ClearAndAbortIfInvalid(ThumbnailView))
            {
                return;
            }
            if (ThumbView == null)
            {
                ThumbView = ThumbnailView as ImmediateLineSeriesBaseView;
                ThumbView.IsThumbnail = true;
                ThumbView.MainView = ImmediateView;
            }

            ThumbView.PrepSurface(surface);
            PrepareFrame(ThumbnailFrame, ThumbView);
            RenderFrame(ThumbnailFrame, ThumbView);
        }
        protected override AxisRange GetRange(Axis axis)
        {
            if (axis != null && axis == XAxis && XColumn != null)
            {
                return new AxisRange(XColumn.Minimum, XColumn.Maximum);
            }

            if (axis != null && axis == YAxis && YColumn != null)
            {
                return new AxisRange(YColumn.Minimum, YColumn.Maximum);
            }

            return null;
        }

        protected override bool ValidateSeries(Rect viewportRect, Rect windowRect, SeriesView view)
        {
            bool isValid = true;

            if (!base.ValidateSeries(viewportRect, windowRect, view)
                || windowRect.IsEmpty
                || viewportRect.IsEmpty
                || XAxis == null
                || YAxis == null
                || XAxis.SeriesViewer == null
                || YAxis.SeriesViewer == null
                || XColumn == null
                || YColumn == null
                || XColumn.Count == 0
                || YColumn.Count == 0
                || FastItemsSource == null
                || FastItemsSource.Count != XColumn.Count
                || FastItemsSource.Count != YColumn.Count
                || XAxis.SeriesViewer == null
                || YAxis.SeriesViewer == null
                || XAxis.ActualMinimumValue == XAxis.ActualMaximumValue
                || YAxis.ActualMinimumValue == YAxis.ActualMaximumValue)
            {
                isValid = false;
            }

            return isValid;
        }

        private double[] _xValues;
        private double[] _yValues;

        public override object GetItem(Point world)
        {
            var index = GetItemIndex(world);

            return index >= 0
                && FastItemsSource != null
                && index < FastItemsSource.Count ? FastItemsSource[index] : null;
        }

        public override int GetItemIndex(Point world)
        {
            int rowIndex = (int)Math.Round(GetExactItemIndex(world));
            return rowIndex;
        }

        public override double GetExactItemIndex(Point world)
        {
            Rect windowRect;
            Rect viewportRect;
            this.View.GetViewInfo(out viewportRect, out windowRect);


            NumericXAxis xAxis = XAxis;
            Rect effectiveViewportRect = this.GetEffectiveViewport(this.View);
            Rect contentViewport = GetContentViewport(View);

            double rowIndex = -1;
            if (XColumn == null)
            {
                return -1;
            }

            if (xAxis != null && !windowRect.IsEmpty && !viewportRect.IsEmpty)
            {
                ScalerParams p = new ScalerParams(windowRect, viewportRect, xAxis.IsInverted, effectiveViewportRect);

                double left = xAxis.GetUnscaledValue(contentViewport.Left, p);
                double right = xAxis.GetUnscaledValue(contentViewport.Right, p);

                double windowX = (world.X - windowRect.Left) / windowRect.Width;
                double xVal = left + (windowX * (right - left));

                int res = XColumn.BinarySearch((v) => xVal.CompareTo(v));

                if (res < 0)
                {
                    res = ~res;
                }
                if (res < 0)
                {
                    return -1;
                }
                if (res > XColumn.Count - 1)
                {
                    return -1;
                }

                double val = XColumn[res];
                if (val == xVal)
                {
                    return res;
                }
                if (val < xVal)
                {
                    if (res == XColumn.Count - 1)
                    {
                        return -1;
                    }
                    double nextVal = XColumn[res + 1];
                    double pr = (xVal - val) / (nextVal - val);
                    return res + pr;
                }
                else
                {
                    if (res == 0)
                    {
                        return -1;
                    }
                    double prevVal = XColumn[res - 1];
                    double pr = (xVal - prevVal) / (val - prevVal);
                    return (res - 1) + pr;
                }
            }

            return rowIndex;
        }

        private Rect GetContentViewport(SeriesView view)
        {
            if (SeriesViewer == null)
            {
                Rect viewport;
                Rect window;
                view.GetViewInfo(out viewport, out window);
                return viewport;
            }
            return SeriesViewer.ContentViewport;
        }

        protected virtual void PrepareFrame(OrderedScatterFrame frame, ImmediateLineSeriesBaseView view)
        {
            double[] xValues;
            double[] yValues;
            if (_xValues == null || _xValues.Length != XColumn.Count)
            {
                _xValues = new double[XColumn.Count];
            }
            if (_yValues == null || _yValues.Length != YColumn.Count)
            {
                _yValues = new double[YColumn.Count];
            }
            xValues = _xValues;
            yValues = _yValues;

            XColumn.CopyTo(xValues, 0);
            YColumn.CopyTo(yValues, 0);

            Rect windowRect;
            Rect viewportRect;
            view.GetViewInfo(out viewportRect, out windowRect);

            ScalerParams xParams = new ScalerParams(
                windowRect,
                viewportRect,
                XAxis.IsInverted);
            xParams.EffectiveViewportRect
                = SeriesViewer.EffectiveViewport;

            ScalerParams yParams = new ScalerParams(
                windowRect,
                viewportRect,
                YAxis.IsInverted);
            yParams.EffectiveViewportRect
                = SeriesViewer.EffectiveViewport;

            XAxis.GetScaledValueList(xValues, 0, xValues.Length, xParams);
            YAxis.GetScaledValueList(yValues, 0, yValues.Length, yParams);

            var viewportWidth = viewportRect.Width;
            var numBuckets = Math.Max(Math.Ceiling(viewportWidth / Resolution), 1);

            var firstVisible = XAxis.GetUnscaledValue(viewportRect.Left, xParams);
            var lastVisible = XAxis.GetUnscaledValue(viewportRect.Right, xParams);

            if (XAxis.IsInverted)
            {
                var swap = firstVisible;
                firstVisible = lastVisible;
                lastVisible = swap;
            }

            var bucketWidth = (lastVisible - firstVisible) / numBuckets;



            var start = XColumn.BinarySearch((xVal) => firstVisible.CompareTo(xVal));
            if (start < 0)
            {
                start = ~start;
                start--;
                if (start < 0)
                {
                    start = 0;
                }
            }
            if (start >= 0 && start < XColumn.Count)
            {
                firstVisible = XColumn[start];
            }

            var end = XColumn.BinarySearch((xVal) => lastVisible.CompareTo(xVal));
            if (end < 0)
            {
                end = ~end;
            }
            if (end > xValues.Length - 1)
            {
                end = xValues.Length - 1;
            }
            if (end >= 0 && end < XColumn.Count)
            {
                lastVisible = XColumn[end];
            }
            bucketWidth = (lastVisible - firstVisible) / numBuckets;



            var x = frame.X;
            var minY = frame.MinY;
            var maxY = frame.MaxY;
            var hasValues = frame.HasValues;

            if (x == null || x.Length != (int)numBuckets)
            {
                x = new double[(int)numBuckets];
                frame.X = x;
            }
            if (minY == null || minY.Length != (int)numBuckets)
            {
                minY = new double[(int)numBuckets];
                frame.MinY = minY;
            }
            if (maxY == null || maxY.Length != (int)numBuckets)
            {
                maxY = new double[(int)numBuckets];
                frame.MaxY = maxY;
            }
            if (hasValues == null || hasValues.Length != (int)numBuckets)
            {
                hasValues = new bool[(int)numBuckets];
                frame.HasValues = hasValues;
            }

            double currMin = double.NaN;
            double currMax = double.NaN;

            bool bucketStarting = true;
            double bucketStartPosition = firstVisible;

            int currentBucket = 0;

            double currentXValue;
            double currentYValue;
            var xColumn = XColumn;
            int inBucket = 0;
            double bucketXValue = 0;

            for (var i = start; i <= end; i++)
            {
                currentXValue = xValues[i];
                currentYValue = yValues[i];

                while (xColumn[i] - bucketStartPosition > bucketWidth)
                {
                    hasValues[currentBucket] = inBucket > 0;
                    x[currentBucket] = bucketXValue; //inBucket > 1 ? (currentBucket / numBuckets) * viewportWidth : currentXValue;
                    minY[currentBucket] = (float)currMin;
                    maxY[currentBucket] = (float)currMax;

                    bucketStarting = true;
                    currMin = double.NaN;
                    currMax = double.NaN;
                    inBucket = 0;
                    bucketStartPosition += bucketWidth;

                    currentBucket++;
                    if (currentBucket > numBuckets - 1)
                    {
                        break;
                    }
                }

                inBucket++;

                if (double.IsNaN(currentYValue))
                {
                    //TODO: alternatively you may want this to create a discontinuity.
                    continue;
                }

                if (bucketStarting)
                {
                    currMin = currentYValue;
                    currMax = currentYValue;
                    bucketXValue = currentXValue;
                    bucketStarting = false;
                }
                else
                {
                    currMin = currMin < currentYValue ? currMin : currentYValue;
                    currMax = currMax > currentYValue ? currMax : currentYValue;
                }
            }

            var last = Math.Min((int)numBuckets - 1, currentBucket);

            x[last] = bucketXValue; // ((double)last / numBuckets) * viewportWidth;
            minY[last] = (float)currMin;
            maxY[last] = (float)currMax;
            hasValues[last] = inBucket > 0;

            for (var i = last + 1; i < numBuckets; i++)
            {
                hasValues[i] = false;
            }
        }

        protected virtual void RenderFrame(OrderedScatterFrame frame, ImmediateLineSeriesBaseView view)
        {
            var bucketCount = frame.X.Length;

            List<SegmentInfo> segments = new List<SegmentInfo>();
            SegmentInfo currentSegment = new SegmentInfo();
            currentSegment.Start = -1;

            double currentMin = double.NaN;
            double currentMax = double.NaN;
            bool currentHasValues = false;
            var x = frame.X;
            var minY = frame.MinY;
            var maxY = frame.MaxY;
            var hasValues = frame.HasValues;
            bool useHairline = UseHairline;

            for (var i = 0; i < bucketCount; i++)
            {
                currentMin = minY[i];
                currentMax = maxY[i];
                currentHasValues = hasValues[i];

                if (double.IsNaN(currentMin) || double.IsNaN(currentMax))
                {
                    if (i - currentSegment.Start > 0 && currentSegment.Start != -1 && currentHasValues)
                    {
                        currentSegment.End = i;
                        segments.Add(currentSegment);
                        currentSegment = new SegmentInfo();
                        currentSegment.Start = -1;
                    }

                    continue;
                }

                if (currentSegment.Start == -1)
                {
                    currentSegment.Start = i;
                }
            }

            if ((bucketCount - 1) - currentSegment.Start > 0
                && currentSegment.Start != -1)
            {
                currentSegment.End = bucketCount - 1;
                segments.Add(currentSegment);
            }

            Rect windowRect;
            Rect viewportRect;
            view.GetViewInfo(out viewportRect, out windowRect);

            double firstVisible = 0;
            double lastVisible = 0;
            if (frame.X.Length > 0)
            {
                firstVisible = frame.X[0];
                lastVisible = frame.X[frame.X.Length - 1];
            }

            Clipper c = GetLineClipper(firstVisible, lastVisible, viewportRect, windowRect);

            view.Rasterize(frame.X, frame.MinY, frame.MaxY, frame.HasValues, segments, c, Resolution, useHairline);

            view.OnRendered();
        }

        internal Clipper GetLineClipper(double firstVisible, double lastVisible, Rect viewportRect, Rect windowRect)
        {
            Clipper clipper = null;

            var left = double.NaN;
            var right = double.NaN;

            if (firstVisible < viewportRect.Left - 2000)
            {
                left = viewportRect.Left - 10;
            }
            if (lastVisible > viewportRect.Right + 2000)
            {
                right = viewportRect.Right + 10;
            }

            clipper = new Clipper(left,
                viewportRect.Bottom + 10, right,
                viewportRect.Top - 10, false);

            return clipper;
        }

        /// <summary>
        /// Overriden in derived classes when they want to respond to the viewport changing.
        /// </summary>
        /// <param name="oldViewportRect">The old viewport rectangle.</param>
        /// <param name="newViewportRect">The new viewport rectangle.</param>
        protected override void ViewportRectChangedOverride(Rect oldViewportRect, Rect newViewportRect)
        {
            RenderSeries(false);
            ImmediateView.OnResize(newViewportRect.Width, newViewportRect.Height);
        }

        /// <summary>
        /// Overriden in derived classes when they want to respond to the chart's window changing.
        /// </summary>
        /// <param name="oldWindowRect">The old window rectangle of the chart.</param>
        /// <param name="newWindowRect">The new window rectangle of the chart.</param>
        protected override void WindowRectChangedOverride(Rect oldWindowRect, Rect newWindowRect)
        {
            RenderSeries(false);
        }

        protected override void DataUpdatedOverride(FastItemsSourceEventAction action, int position, int count, string propertyName)
        {
            base.DataUpdatedOverride(action, position, count, propertyName);

            RenderSeries(false);
        }
    }

    public class SegmentInfo
    {
        public int Start;
        public int End;

        public List<Point> MinPoints;
        public List<Point> MaxPoints;
    }

    public class OrderedScatterFrame
    {
        public double[] X { get; set; }
        public double[] MinY { get; set; }
        public double[] MaxY { get; set; }
        public bool[] HasValues { get; set; }
    }
}
